#!/bin/sh
#
# Compute delta between source and destination file, using a special strategy for
# files in ZIP format, including ZIP, JAR, WAR and EAR files. Very efficient for
# developing patches for Java software.
#
# Usage: zd.sh [-b|-c] <recursion depth> <source> <dest> <delta>
#
# Notes:
#   - BOM = Bill Of Materials (list of files with hash and relative path)
#   - we cannot use file extension to determine its type (no extension in tmp file names)
#   - to test if file is a JAR we look for .class files inside it
#
# modes
# 0: binary delta only
# 1: zd delta on zip files except jars, no recurse
# 2: zd delta on all zip files, no recurse
# 3: zd delta on zip files except jars + recurse
# 4: zd delta on all zip files + recurse
#

ZDOPT="-b"
ZDBINOPT=
while getopts bc option
do	
	case "$option" in
	b)	ZDOPT="-b"
                ZDBINOPT=
		;;
	c)	ZDOPT="-c"
                ZDBINOPT="-c"
		;;
	esac
done
shift `expr $OPTIND - 1`

PROGNAME=`basename $0`

if [ $# -ne 4 ]; then
    echo "$PROGNAME is a command line tool that computes delta between source and destination file,"
    echo "using a special strategy for files in ZIP format, including ZIP, JAR, WAR and EAR files."
	echo "Very efficient for developing patches for Java software."
	echo
    echo "usage: $PROGNAME [-b|-c] <mode> <source> <dest> <delta>"
    echo "\t-b: compute a binary delta (default)"
    echo "\t-c: copy mode: <delta> is equal to <dest> file"
    exit 1;
fi

MODE=$1
SOURCE=$2
DEST=$3
DELTAFILE=$4

LVL="+$LVL" ; export LVL # indent logging when recursing

COUNTERFILE=/tmp/counter

###############
## FUNCTIONS ##
###############

#-----------------------------------
#	log
#-----------------------------------
log()
{
	printf "`date "+%Y-%m-%d %H:%M:%S"` $LVL $*"
}

#-----------------------------------
#	incrementCounter
#-----------------------------------
incrementCounter()
{
    if [ -f $COUNTERFILE ];
    then
        value=`grep -m 1 "^$1=[0-9]*" $COUNTERFILE | cut -d= -f2`
        if [ -n "$value" ];
        then
            sed -e "s/\(^$1=\)[0-9]*/\\1`expr $value + 1`/" $COUNTERFILE > ${COUNTERFILE}tmp; mv ${COUNTERFILE}tmp ${COUNTERFILE}
        else
            echo "$1=1" >> $COUNTERFILE
        fi
    else
        echo "$1=1" > $COUNTERFILE
    fi
}

##########
## MAIN ##
##########

# check if unzipable (status 0 => unzipable)
zipinfo -h "$SOURCE" > /dev/null 2>&1 ; STS=$? 
zipinfo -h "$DEST" > /dev/null 2>&1 ; STD=$?

if [ $MODE -eq 1 -o $MODE -eq 3 ] # mode 1 and mode 3 exclude jar files
then
	# check if JAR file (set status to 99 if contains a .class file)
	if [ $STS -eq 0 ] ; then if zipinfo -1 $SOURCE | fgrep -qs -m 1 '.class'; then STS=99; fi ; fi
	if [ $STD -eq 0 ] ; then if zipinfo -1 $DEST | fgrep -qs -m 1 '.class'; then STD=99; fi ; fi
fi

if [ $MODE -eq 0 -o $STS -ne 0 -o $STD -ne 0 ] # mode 0 ignores unzipable files
then # one if the files is not unzipable or is a jar
	log "create binary delta (OPT=$ZDBINOPT) `basename $DELTAFILE` ..."
	incrementCounter COUNTPATCH
	zdbin.sh -e $ZDBINOPT "$SOURCE" "$DEST" "$DELTAFILE"
	echo " DONE"
	exit 0
fi

# both files are unzipable

TMPDIR=`mktemp -d` # folder to store deltas
TMPS=`mktemp` ; TMPD=`mktemp` # extracted source and destination files
BOMS=$TMPDIR/source.bom ; BOMD=$TMPDIR/dest.bom # source and destination BOM

# compute BOMs
log "create source BOM..."      ; zdbom.sh "$SOURCE" | sort > $BOMS ; echo " DONE `cat $BOMS | wc -l` files"
log "create destination BOM..." ; zdbom.sh "$DEST" | sort > $BOMD ; echo " DONE `cat $BOMD | wc -l` files"

# strip  BOMs
log "strip BOMS..."
comm -23 $BOMS $BOMD > $BOMS.s # unique in source BOM
comm -13 $BOMS $BOMD > $BOMD.s # unique in destination BOM
# add missing items to stripped source BOM
# for case: file exists in same location in source and destination, and is duplicated in destination
# Add entry into source and dest BOM, else it would be seen as a move scenario
join -v 2 $BOMS.s $BOMD.s | while read HASH FPATH; do grep "^$HASH " $BOMS ; done > $BOMS.x
cat $BOMS.s $BOMS.x | sort -u > $BOMS.sx ; mv $BOMS.sx $BOMS.s ; mv $BOMS.s $BOMS
cat $BOMD.s $BOMS.x | sort -u > $BOMD.sx ; mv $BOMD.sx $BOMD.s ; mv $BOMD.s $BOMD
rm $BOMS.x 
echo " DONE"

cp $BOMD $BOMD.l

# count removed files
cat $BOMS | while read HASH FILEPATH
do
	[ -z "`grep \" $FILEPATH$\" $BOMD`" ] && {
		CNTHD=`grep -c "^$HASH " $BOMD`
		CNTHS=`grep -c "^$HASH " $BOMS`
		[ "$CNTHD" != "1" -o "$CNTHS" != "1" ] && {
			log "[REMOVE] $FILEPATH\n"
			incrementCounter COUNTREMOVE
		}
	}
done

# start computing delta
#join -v 2 $BOMS $BOMD | while read HASH FILEPATH # files in destination but not in source
cat $BOMD | while read HASH FILEPATH # files in destination but not in source, plus duplicate
do
	# same entry in source and destination, continue
	[ -n "`grep \"^$HASH $FILEPATH$\" $BOMS`" ] && {
		continue
	}
	
	NDXHD=`grep -m 1 -n "^$HASH " "$BOMD.l" | sed 's/:.*$//'` # index of same hash in destination BOM
	NDXPS=`grep -m 1 -n "^.* $FILEPATH$" "$BOMS" | sed 's/:.*$//'` # index of same path in source BOM
	
	# 'disable' current line in BOMD.l so index not used twice
	awk -v idxd=$NDXHD '{ if (NR==idxd) print "-"; else print $0 }' $BOMD.l > $BOMD.a
	rm $BOMD.l; mv $BOMD.a $BOMD.l

	if [ -z "`grep -m 1 \"^$HASH \" $BOMS`" ] ; then # if hash not in source
		if [ -n "$NDXPS" ]; then # if path in source
			log "[PATCH] $FILEPATH\n"
			log "compute delta."
			unzip -p "$SOURCE" "$FILEPATH" > $TMPS ; printf "."
			unzip -p "$DEST" "$FILEPATH" > $TMPD ; echo "."
			if [ $MODE -eq 1 -o $MODE -eq 2 ] # mode 1 and mode 2 exclude recursion
			then
				incrementCounter COUNTPATCH
				zdbin.sh -e $ZDBINOPT $TMPS $TMPD $TMPDIR/${NDXPS}-${NDXHD}
			else
				zd.sh $ZDOPT $MODE $TMPS $TMPD $TMPDIR/${NDXPS}-${NDXHD}
			fi
		else # if path not in source
			log "[ADD] $FILEPATH\n"
			unzip -p "$DEST" "$FILEPATH" > $TMPDIR/${NDXHD}.new ;
			incrementCounter COUNTADD
		fi
	else
		CNTHD=`grep -c "^$HASH " $BOMD`
		CNTHS=`grep -c "^$HASH " $BOMS`
		if [ "$CNTHD" = "1" -a "$CNTHS" = "1" ] ; then
			log "[MOVE] $FILEPATH\n"
			incrementCounter COUNTMOVE
		else
			log "[COPY] $FILEPATH\n"
			incrementCounter COUNTCOPY
		fi
	fi
done

rm $TMPS $TMPD $BOMD.l

if ls -1 $TMPDIR | grep -qc -m 1 "^[0-9]" ; then # folder contains at least one delta file (starting with a digit)
	log "create delta file `basename ${DELTAFILE}.zip`..." ; zip -q -9 -j ${DELTAFILE}.zip -r $TMPDIR ; echo " DONE"
else
	log "no delta files\n"
fi

rm -rf $TMPDIR

exit 1
